import csv
import requests
import time

from abc import ABC, abstractmethod
from typing import Any, Dict, List, Tuple

try:
    import aiohttp
    import asyncio
    ASYNC_AVAILABLE = True
except ImportError:
    ASYNC_AVAILABLE = False


class UnauthorizedException(Exception):

    def __init__(self, message: str = 'Unauthorized'):
        super().__init__(message)


class BaseNextBrain(ABC):

    def __init__(
        self,
        access_token: str,
        backend_url: str = 'https://api.nextbrain.ai',
        is_app: bool = False,
    ):
        self.access_token = access_token
        self.backend_url = backend_url
        self.is_app = is_app

    @staticmethod
    def load_csv(path: str) -> List[List[Any]]:
        with open(path, 'r') as read_obj:
            return list(csv.reader(read_obj))

    @abstractmethod
    def wait_model(self, model_id: str, wait_imported: bool = True):
        raise NotImplementedError()

    @abstractmethod
    def upload_model(self, table: List[List[Any]]):
        raise NotImplementedError()

    @abstractmethod
    def train_model(self, model_id: str, target: str, is_lightning: bool = False):
        raise NotImplementedError()

    @abstractmethod
    def predict_model(self, model_id: str, header: List[str], rows: List[List[Any]]) -> List[Any]:
        raise NotImplementedError()

    @abstractmethod
    def random_predict(self, model_id: str) -> List[Any]:
        raise NotImplementedError()

    @abstractmethod
    def get_predict_columns(self, model_id: str) -> List[str]:
        raise NotImplementedError()

    @abstractmethod
    def upload_and_predict(self, table: List[List[Any]], predict_table: List[List[Any]], target: str, is_lightning: bool = False) -> Dict:
        raise NotImplementedError()


class NextBrain(BaseNextBrain):

    def __init__(self, access_token: str, *args, **kwargs):
        super().__init__(access_token, *args, **kwargs)

    def get_accuracy(self, model_id: str) -> None:
        if self.is_app:
            response = requests.get(
                f'{self.backend_url}/app/acc/{model_id}',
                headers={
                    'access_token': self.access_token
                }
            )
        else:
            response = requests.post(f'{self.backend_url}/model/acc_token/{model_id}', json={
                'access_token': self.access_token,
            })

        if response.status_code == 401:
            raise UnauthorizedException()

        return response.json()

    def wait_model(self, model_id: str, wait_imported: bool = True) -> None:
        while True:
            if self.is_app:
                response = requests.post(
                    f'{self.backend_url}/app/status/{model_id}',
                    headers={
                        'access_token': self.access_token
                    }
                )
            else:
                response = requests.post(f'{self.backend_url}/model/status_token/{model_id}', json={
                    'access_token': self.access_token,
                })

            if response.status_code == 401:
                raise UnauthorizedException()

            data = response.json()
            if wait_imported:
                if data['dataset_status'] == 'imported':
                    return
                elif data['dataset_status'] == 'error':
                    raise Exception('Error importing model')

            else:
                if data['status'] == 'trained':
                    return
                elif data['status'] == 'error':
                    raise Exception('Error training model')

            time.sleep(2)

    def upload_model(self, table: List[List[Any]]) -> str:
        if self.is_app:
            response = requests.post(
                f'{self.backend_url}/app/import_matrix',
                json={
                    'matrix': table,
                },
                headers={
                    'access_token': self.access_token
                }
            )
        else:
            response = requests.post(
                f'{self.backend_url}/csv/import_matrix_token',
                json={
                    'access_token': self.access_token,
                    'matrix': table,
                }
            )

        if response.status_code == 401:
            raise UnauthorizedException()

        data = response.json()
        model_id = data['model']['id']
        self.wait_model(model_id)
        return model_id

    def train_model(self, model_id: str, target: str, is_lightning: bool = False) -> None:
        if self.is_app:
            response = requests.post(
                f'{self.backend_url}/app/train',
                json={
                    'target': target,
                    'model_id': model_id,
                    'is_lightning': is_lightning,
                },
                headers={
                    'access_token': self.access_token
                }
            )
        else:
            response = requests.post(f'{self.backend_url}/model/train_token', json={
                'access_token': self.access_token,
                'target': target,
                'model_id': model_id,
                'is_lightning': is_lightning,
            })

        if response.status_code == 401:
            raise UnauthorizedException()

        self.wait_model(model_id, wait_imported=False)

    def predict_model(self, model_id: str, table: List[List[Any]]) -> Dict:
        header = table[0]
        rows = table[1:]
        if self.is_app:
            response = requests.post(
                f'{self.backend_url}/app/predict/{model_id}',
                json={
                    'header': header,
                    'rows': rows,
                },
                headers={
                    'access_token': self.access_token
                }
            )
        else:
            response = requests.post(f'{self.backend_url}/model/predict_token/{model_id}',
                                     json={
                                         'access_token': self.access_token,
                                         'header': header,
                                         'rows': rows,
                                     }
                                     )

        if response.status_code == 401:
            raise UnauthorizedException()

        return response.json()

    def random_predict(self, model_id: str) -> Dict:
        if self.is_app:
            response = requests.post(
                f'{self.backend_url}/app/random_predict/{model_id}',
                headers={
                    'access_token': self.access_token
                }
            )
        else:
            response = requests.post(
                f'{self.backend_url}/model/random_predict_token/{model_id}',
                json={
                    'access_token': self.access_token,
                }
            )

        if response.status_code == 401:
            raise UnauthorizedException()

        return response.json()

    def get_predict_columns(self, model_id: str) -> List[str]:
        if self.is_app:
            response = requests.get(
                f'{self.backend_url}/app/predict_columns/{model_id}',
                headers={
                    'access_token': self.access_token
                }
            )
        else:
            response = requests.post(
                f'{self.backend_url}/model/predict_columns_token/{model_id}',
                json={
                    'access_token': self.access_token,
                }
            )

        if response.status_code == 401:
            raise UnauthorizedException()

        return response.json()

    def upload_and_predict(self, table: List[List[Any]], predict_table: List[List[Any]], target: str, is_lightning: bool = False) -> Tuple[str, Dict]:
        model_id = self.upload_model(table)
        self.train_model(model_id, target, is_lightning=is_lightning)
        return model_id, self.predict_model(model_id, predict_table)

    def delete_model(self, model_id: str) -> None:
        if self.is_app:
            response = requests.post(
                f'{self.backend_url}/app/delete_model/{model_id}',
                headers={
                    'access_token': self.access_token
                }
            )
        else:
            response = requests.post(f'{self.backend_url}/model/delete_model_token/{model_id}',
                                     json={
                                         'access_token': self.access_token,
                                     }
                                     )

        if response.status_code == 401:
            raise UnauthorizedException()


class AsyncNextBrain(BaseNextBrain):

    def __init__(self, access_token: str, *args, **kwargs):
        if not ASYNC_AVAILABLE:
            raise ImportError(
                'asyncio and aiohttp are required for async usage')
        super().__init__(access_token, *args, **kwargs)

    async def get_accuracy(self, model_id: str) -> float:
        async with aiohttp.ClientSession() as session:
            if self.is_app:
                response = await session.get(
                    f'{self.backend_url}/app/acc/{model_id}',
                    headers={
                        'access_token': self.access_token
                    }
                )
            else:
                response = await session.post(f'{self.backend_url}/model/acc_token/{model_id}', json={
                    'access_token': self.access_token,
                })

            if response.status == 401:
                raise UnauthorizedException()

            return await response.json()

    async def wait_model(self, model_id: str, wait_imported: bool = True) -> None:
        async with aiohttp.ClientSession() as session:
            while True:
                kwargs = {}
                if self.is_app:
                    url = f'{self.backend_url}/app/status/{model_id}'
                    kwargs['headers'] = {
                        'access_token': self.access_token
                    }
                else:
                    url = f'{self.backend_url}/model/status_token/{model_id}'
                    kwargs['json'] = {
                        'access_token': self.access_token
                    }

                async with session.post(url, **kwargs) as response:
                    if response.status == 401:
                        raise UnauthorizedException()
                    data = await response.json()
                    if wait_imported:
                        if data['dataset_status'] == 'imported':
                            return
                        elif data['dataset_status'] == 'error':
                            raise Exception('Error importing model')

                    else:
                        if data['status'] == 'trained':
                            return
                        elif data['status'] == 'error':
                            raise Exception('Error training model')

                await asyncio.sleep(2)

    async def upload_model(self, table: List[List[Any]]) -> str:
        kwargs = {}
        if self.is_app:
            url = f'{self.backend_url}/app/import_matrix'
            kwargs['json'] = {
                'matrix': table,
            }
            kwargs['headers'] = {
                'access_token': self.access_token
            }
        else:
            url = f'{self.backend_url}/csv/import_matrix_token'
            kwargs['json'] = {
                'matrix': table,
                'access_token': self.access_token
            }

        async with aiohttp.ClientSession() as session:
            async with session.post(url, **kwargs) as response:
                data = await response.json()
                model_id = data['model']['id']
                await self.wait_model(model_id)
                return model_id

    async def train_model(self, model_id: str, target: str, is_lightning: bool = False) -> None:
        json = {
            'target': target,
            'model_id': model_id,
            'is_lightning': is_lightning,
        }

        if self.is_app:
            url = f'{self.backend_url}/app/train'
            kwargs = {
                'json': json,
                'headers': {
                    'access_token': self.access_token
                }
            }
        else:
            url = f'{self.backend_url}/model/train_token'
            json['access_token'] = self.access_token
            kwargs = {
                'json': json
            }

        async with aiohttp.ClientSession() as session:
            async with session.post(url, **kwargs) as response:
                # Sequential await is required
                await response.json()
                await self.wait_model(model_id, wait_imported=False)

    async def predict_model(self, model_id: str, table: List[List[Any]]) -> Dict:
        header = table[0]
        rows = table[1:]
        json = {
            'header': header,
            'rows': rows,
        }

        if self.is_app:
            url = f'{self.backend_url}/app/predict/{model_id}'
            kwargs = {
                'json': json,
                'headers': {
                    'access_token': self.access_token
                }
            }
        else:
            url = f'{self.backend_url}/model/predict_token/{model_id}'
            json['access_token'] = self.access_token
            kwargs = {
                'json': json
            }

        async with aiohttp.ClientSession() as session:
            async with session.post(url, **kwargs) as response:
                return await response.json()

    async def random_predict(self, model_id: str) -> Dict:
        if self.is_app:
            url = f'{self.backend_url}/app/random_predict/{model_id}'
            kwargs = {
                'headers': {
                    'access_token': self.access_token
                }
            }
        else:
            url = f'{self.backend_url}/model/random_predict_token/{model_id}'
            kwargs = {
                'json': {
                    'access_token': self.access_token
                }
            }

        async with aiohttp.ClientSession() as session:
            async with session.post(url, **kwargs) as response:
                return await response.json()

    async def get_predict_columns(self, model_id: str) -> List[str]:
        if self.is_app:
            url = f'{self.backend_url}/app/predict_columns/{model_id}'
            kwargs = {
                'headers': {
                    'access_token': self.access_token
                }
            }
        else:
            url = f'{self.backend_url}/model/predict_columns_token/{model_id}'
            kwargs = {
                'json': {
                    'access_token': self.access_token
                }
            }

        async with aiohttp.ClientSession() as session:
            method = session.get if self.is_app else session.post
            async with method(url, **kwargs) as response:
                return await response.json()

    async def upload_and_predict(self, table: List[List[Any]], predict_table: List[List[Any]], target: str, is_lightning: bool = False) -> Tuple[str, Dict]:
        # Required sequential await
        model_id = await self.upload_model(table)
        await self.train_model(model_id, target, is_lightning=is_lightning)
        return model_id, await self.predict_model(model_id, predict_table)

    async def delete_model(self, model_id: str) -> None:
        if self.is_app:
            url = f'{self.backend_url}/app/delete_model/{model_id}'
            kwargs = {
                'headers': {
                    'access_token': self.access_token
                }
            }
        else:
            url = f'{self.backend_url}/model/delete_model_token/{model_id}'
            kwargs = {
                'json': {
                    'access_token': self.access_token,
                }
            }

        async with aiohttp.ClientSession() as session:
            async with session.post(url, **kwargs) as response:
                if response.status == 401:
                    raise UnauthorizedException()
                # Sequential await is required
                await response.json()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nextbrain']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'nextbrain',
    'version': '1.0.0',
    'description': 'Convenient access to the NextBrain API from python',
    'long_description': "# NextBrain AI\nConvenient access to the [NextBrain AI](https://nextbrain.ai) API from python\n\n## Installation\n```bash\npip install nextbrain\n```\n\nIf you want to use the async version you need to install `asyncio` and `aiohttp`:\n\n```bash\npip install asyncio aiohttp\n```\n\n## Normal usage\n\n### All steps in one.\n```python\nfrom nextbrain import NextBrain\nfrom typing import Any, List\n\ndef main():\n    nb = NextBrain('<YOUR-ACCESS-TOKEN-HERE>')\n\n    # You can create your custom table and predict table by your own from any source\n    # It is a list of list, where the first row contains the header\n    # Example:\n    # [\n    #   [ Column1, Column2, Column3 ],\n    #   [       1,       2,       3 ],\n    #   [       4,       5,       6 ]\n    # ]\n    table: List[List[Any]] = nb.load_csv('<PATH-TO-YOUR-TRAINING-CSV>')\n    predict_table: List[List[Any]] = nb.load_csv('<PATH-TO-YOUR-PREDICTING-CSV>')\n\n    model_id, response = nb.upload_and_predict(table, predict_table, '<YOUR-TARGET-COLUMN>')\n    # model_id is also returned in order to predict multiple times against same model\n    print(response)\n\nif __name__ == '__main__':\n    main()\n```\n\n### Step by step\n```python\nfrom nextbrain import NextBrain\nfrom typing import Any, List\n\ndef main():\n    nb = NextBrain('<YOUR-ACCESS-TOKEN-HERE>')\n\n    # You can create your custom table and predict table by your own from any source\n    table: List[List[Any]] = nb.load_csv('<PATH-TO-YOUR-TRAINING-CSV>')\n    # Upload the model to NextBrain service\n    model_id: str = nb.upload_model(table)\n    # Train the model\n    # You can re-train a previous model\n    nb.train_model(model_id, '<YOUR-TARGET-COLUMN>')\n\n    predict_table: List[List[Any]] = nb.load_csv('<PATH-TO-YOUR-PREDICTING-CSV>')\n    # You can predict multiple using the same model (don't need to create a new model each time)\n    response = nb.predict_model(model_id, predict_table)\n    print(response)\n\nif __name__ == '__main__':\n    main()\n```\n\n## Async usage\n\n### All steps in one.\n```python\nfrom nextbrain import AsyncNextBrain\nfrom typing import Any, List\n\nasync def main():\n    nb = AsyncNextBrain('<YOUR-ACCESS-TOKEN-HERE>')\n\n    # You can create your custom table and predict table by your own from any source\n    table: List[List[Any]] = nb.load_csv('<PATH-TO-YOUR-TRAINING-CSV>')\n    predict_table: List[List[Any]] = nb.load_csv('<PATH-TO-YOUR-PREDICTING-CSV>')\n\n    model_id, response = await nb.upload_and_predict(table, predict_table, '<YOUR-TARGET-COLUMN>')\n    # model_id is also returned in order to predict multiple times against same model\n    print(response)\n\nif __name__ == '__main__':\n    import asyncio\n    asyncio.run(main())\n```\n\n### Step by step\n```python\nfrom nextbrain import AsyncNextBrain\nfrom typing import Any, List\n\nasync def main():\n    nb = AsyncNextBrain('<YOUR-ACCESS-TOKEN-HERE>')\n\n    # You can create your custom table and predict table by your own from any source\n    table: List[List[Any]] = nb.load_csv('<PATH-TO-YOUR-TRAINING-CSV>')\n    # Upload the model to NextBrain service\n    model_id: str = await nb.upload_model(table)\n    # Train the model\n    # You can re-train a previous model\n    await nb.train_model(model_id, '<YOUR-TARGET-COLUMN>')\n\n    predict_table: List[List[Any]] = nb.load_csv('<PATH-TO-YOUR-PREDICTING-CSV>')\n    # You can predict multiple using the same model (don't need to create a new model each time)\n    response = await nb.predict_model(model_id, predict_table)\n    print(response)\n\nif __name__ == '__main__':\n    import asyncio\n    asyncio.run(main())\n```\n\n## Extra notes\n\nEverytime you train, you can select an option to create lightning models. `is_lightning` is an optional parameter that by default is set to `False` but can be overrided in `train_model` and `upload_and_predict`.\n\nWe also recommend that you investigate all the methods that the class provides you with to make the most of the functionalities we offer. For example, you can use the `get_accuracy` method to obtain all the information about the performance of your model.\n",
    'author': 'Softpoint Consultores S.L.',
    'author_email': 'info@softpoint.es',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/NextBrain-ai/nextbrain-python',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
